# Generated by Django 4.2.13 on 2024-05-27 10:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0106_remove_homestationfilter_constellations_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='homestationfilter',
            old_name='structures',
            new_name='evelocation',
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-12 11:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0101_corporatecontract_corporatecontractitem'),
    ]

    operations = [
        migrations.AddField(
            model_name='corporationaudit',
            name='last_change_contracts',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='corporationaudit',
            name='last_update_contracts',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
    ]



# Generated by Django 2.2.12 on 2020-09-01 05:37

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0031_auto_20200828_0855'),
    ]

    operations = [
        migrations.AddField(
            model_name='charactermarketorder',
            name='region_name',
            field=models.ForeignKey(
                default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='corptools.MapRegion'),
        ),
        migrations.AddField(
            model_name='charactermarketorder',
            name='type_name',
            field=models.ForeignKey(
                default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='corptools.EveItemType'),
        ),
        migrations.AddField(
            model_name='corporationmarketorder',
            name='region_name',
            field=models.ForeignKey(
                default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='corptools.MapRegion'),
        ),
        migrations.AddField(
            model_name='corporationmarketorder',
            name='type_name',
            field=models.ForeignKey(
                default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='corptools.EveItemType'),
        ),
    ]



# Generated by Django 4.0.10 on 2023-05-22 09:38

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0085_characteraudit_last_update_mining_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='CharacterIndustryJob',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('activity_id', models.IntegerField()),
                ('blueprint_id', models.BigIntegerField()),
                ('blueprint_location_id', models.BigIntegerField()),
                ('blueprint_type_id', models.BigIntegerField()),
                ('completed_character_id', models.BigIntegerField(
                    blank=True, default=None, null=True)),
                ('completed_date', models.DateTimeField(
                    blank=True, default=None, null=True)),
                ('cost', models.DecimalField(blank=True, decimal_places=2,
                 default=None, max_digits=20, null=True)),
                ('duration', models.IntegerField()),
                ('end_date', models.IntegerField()),
                ('facility_id', models.BigIntegerField()),
                ('installer_id', models.IntegerField()),
                ('job_id', models.IntegerField()),
                ('licensed_runs', models.IntegerField(
                    blank=True, default=None, null=True)),
                ('output_location_id', models.BigIntegerField()),
                ('pause_date', models.DateTimeField(
                    blank=True, default=None, null=True)),
                ('probability', models.FloatField(
                    blank=True, default=None, null=True)),
                ('product_type_id', models.IntegerField()),
                ('runs', models.IntegerField()),
                ('start_date', models.DateTimeField()),
                ('station_id', models.BigIntegerField()),
                ('status', models.CharField(max_length=15)),
                ('successful_runs', models.IntegerField(
                    blank=True, default=None, null=True)),
                ('blueprint_type_name', models.ForeignKey(default=None, null=True,
                 on_delete=django.db.models.deletion.SET_NULL, related_name='blueprint_type', to='corptools.eveitemtype')),
                ('character', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.characteraudit')),
                ('product_type_name', models.ForeignKey(default=None, null=True,
                 on_delete=django.db.models.deletion.SET_NULL, related_name='product_type', to='corptools.eveitemtype')),
            ],
        ),
    ]



# Generated by Django 4.0.10 on 2023-11-25 08:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0095_lastloginfilter_main_corp_only_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='rolefilter',
            name='main_only',
            field=models.BooleanField(default=False),
        ),
    ]



import time

from bravado.exception import HTTPNotModified

from django.core.cache import cache

from allianceauth.services.hooks import get_extension_logger

MAX_ETAG_LIFE = 60 * 60 * 24 * 7  # 7 Days

logger = get_extension_logger(__name__)


class NotModifiedError(Exception):
    pass


def get_etag_key(operation):
    return "etag-" + operation._cache_key()


def get_etag_header(operation):
    return cache.get(get_etag_key(operation), False)


def del_etag_header(operation):
    return cache.delete(get_etag_key(operation), False)


def inject_etag_header(operation):
    etag = get_etag_header(operation)
    logger.debug(
        f"ETag: get_etag {operation.operation.operation_id} - {stringify_params(operation)} - etag:{etag}")
    if etag:
        operation.future.request.headers["If-None-Match"] = etag


def rem_etag_header(operation):
    logger.debug(
        f"ETag: rem_etag {operation.operation.operation_id} - {stringify_params(operation)}")
    if "If-None-Match" in operation.future.request.headers:
        del operation.future.request.headers["If-None-Match"]


def set_etag_header(operation, headers):
    etag_key = get_etag_key(operation)
    etag = headers.headers.get('ETag', None)
    if etag is not None:
        result = cache.set(etag_key, etag, MAX_ETAG_LIFE)
        logger.debug(
            f"ETag: set_etag {operation.operation.operation_id} - {stringify_params(operation)} - etag:{etag} - stored:{result}")


def stringify_params(operation):
    out = []
    for p, v in operation.future.request.params.items():
        out.append(f"{p}: {v}")
    return ", ".join(out)


def etag_results(operation, token, force_refresh=False, disable_verification=False):
    _start_tm = time.perf_counter()
    results = list()
    # override to always get the raw response for expiry header
    # print("******************", operation.operation.swagger_spec.config["validate_responses"])

    operation.request_config.also_return_response = True

    try:
        logger.debug(f"ETAG Validate Pre: {operation.operation.swagger_spec.config['validate_responses']}")
    except Exception:
        logger.warning("Pre operation.operation.swagger_spec.config['validate_responses'] not checkable")

    if disable_verification:
        operation.operation.swagger_spec.config["validate_responses"] = False
    else:
        operation.operation.swagger_spec.config["validate_responses"] = True

    try:
        logger.debug(f"ETAG Validate Override: {operation.operation.swagger_spec.config['validate_responses']}")
    except Exception:
        logger.warning("Override operation.operation.swagger_spec.config['validate_responses'] not checkable")

    if token:
        operation.future.request.headers["Authorization"] = "Bearer " + \
            token.valid_access_token()

    # Global try to enforce validation reset after we are done regardless of outcome here.
    try:
        if "page" in operation.operation.params:
            logger.debug(
                f"ETag: Pages Start {operation.operation.operation_id} - {stringify_params(operation)}")
            current_page = 1
            total_pages = 1
            etags_incomplete = False

            # loop all pages and add data to output array
            while current_page <= total_pages:
                _pg_tm = time.perf_counter()
                operation.future.request.params["page"] = current_page
                # will use cache if applicable
                try:
                    if not etags_incomplete and not force_refresh:
                        logger.debug(
                            f"ETag: Injecting Header {operation.operation.operation_id} - {stringify_params(operation)}")
                        inject_etag_header(operation)
                    else:
                        logger.debug(
                            f"ETag: Removing Header {operation.operation.operation_id} F:{force_refresh} Ei:{etags_incomplete} - {stringify_params(operation)}")
                        rem_etag_header(operation)

                    result, headers = operation.result()
                    total_pages = int(headers.headers['X-Pages'])
                    logger.warning(get_etag_header(operation))
                    logger.warning(headers.headers.get('ETag'))

                    if get_etag_header(operation) == headers.headers.get('ETag') and not force_refresh and not etags_incomplete:
                        # if django esi is returning our cache check it manualy.
                        raise NotModifiedError()

                    if force_refresh:
                        logger.debug(
                            f"ETag: Removing Etag {operation.operation.operation_id} F:{force_refresh} - {stringify_params(operation)}")
                        del_etag_header(operation)
                    else:
                        logger.debug(
                            f"ETag: Saving Etag {operation.operation.operation_id} F:{force_refresh} - {stringify_params(operation)}")
                        set_etag_header(operation, headers)

                    # append to results list to be seamless to the client
                    results += result
                    current_page += 1

                    if not etags_incomplete and not force_refresh:
                        logger.debug(
                            f"ETag: No Etag {operation.operation.operation_id} - {stringify_params(operation)}")
                        current_page = 1  # reset to page 1 and fetch everything
                        results = list()
                        etags_incomplete = True

                except (HTTPNotModified) as e:  # etag is match from ESI
                    logger.debug(
                        f"ETag: HTTPNotModified Hit ETag {operation.operation.operation_id} Ei:{etags_incomplete} - {stringify_params(operation)} - P:{e.response.headers['X-Pages']}")
                    total_pages = int(e.response.headers['X-Pages'])

                    if not etags_incomplete:
                        current_page += 1
                    else:
                        current_page = 1  # reset to page 1 and fetch everything, we should not get here
                        results = list()

                except NotModifiedError:  # etag is match in cache
                    logger.debug(
                        f"ESI_TIME: PAGE {time.perf_counter() - _pg_tm} {operation.operation.operation_id} {stringify_params(operation)}")
                    total_pages = int(headers.headers['X-Pages'])

                    if not etags_incomplete:
                        current_page += 1
                    else:
                        current_page = 1  # reset to page 1 and fetch everything, we should not get here
                        results = list()

                logger.debug(
                    f"ETag: No Etag {operation.operation.operation_id} - {stringify_params(operation)}")

            if not etags_incomplete and not force_refresh:
                raise NotModifiedError()

        else:  # it doesn't so just return as usual
            if not force_refresh:
                inject_etag_header(operation)
            try:
                results, headers = operation.result()
            except HTTPNotModified as e:
                logger.debug(
                    f"ETag: HTTPNotModified Hit ETag {operation.operation.operation_id} - {stringify_params(operation)}")
                set_etag_header(operation, e.response)
                raise NotModifiedError()

            if get_etag_header(operation) == headers.headers.get('ETag') and not force_refresh:
                # etag is match in cache
                logger.debug(
                    f"ETag: result Cache Hit ETag {operation.operation.operation_id} - {stringify_params(operation)}")
                set_etag_header(operation, headers)
                raise NotModifiedError()

            if force_refresh:
                logger.debug(
                    f"ETag: Removing Etag {operation.operation.operation_id} F:{force_refresh} - {stringify_params(operation)}")
                del_etag_header(operation)
            else:
                logger.debug(
                    f"ETag: Saving Etag {operation.operation.operation_id} F:{force_refresh} - {stringify_params(operation)}")
                set_etag_header(operation, headers)
    except Exception:
        # Re-raise everything
        raise
    finally:
        # reset the validation.
        if not disable_verification:
            operation.operation.swagger_spec.config["validate_responses"] = True
        try:
            logger.debug(f"ETAG Validate Post: {operation.operation.swagger_spec.config['validate_responses']}")
        except Exception:
            logger.warning("Post operation.operation.swagger_spec.config['validate_responses'] not checkable")

    logger.debug(
        f"ESI_TIME: OVERALL {time.perf_counter() - _start_tm} {operation.operation.operation_id} {stringify_params(operation)}")
    return results



# Generated by Django 4.0.7 on 2022-09-18 07:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0076_remove_characterasset_corptools_c_type_id_05ccc2_idx_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='fullyloadedfilter',
            name='reversed_logic',
            field=models.BooleanField(
                default=False, help_text='If set all members WITHOUT audit fully loaded will pass the test.'),
        ),
    ]



# Generated by Django 4.0.7 on 2022-10-05 05:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0077_fullyloadedfilter_reversed_logic'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='characterwalletjournalentry',
            index=models.Index(
                fields=['date'], name='corptools_c_date_aebcea_idx'),
        ),
        migrations.AddIndex(
            model_name='characterwalletjournalentry',
            index=models.Index(fields=['entry_id'],
                               name='corptools_c_entry_i_d9e191_idx'),
        ),
        migrations.AddIndex(
            model_name='characterwalletjournalentry',
            index=models.Index(fields=['ref_type'],
                               name='corptools_c_ref_typ_fe0ec9_idx'),
        ),
        migrations.AddIndex(
            model_name='corporationwalletjournalentry',
            index=models.Index(
                fields=['date'], name='corptools_c_date_ca15ac_idx'),
        ),
        migrations.AddIndex(
            model_name='corporationwalletjournalentry',
            index=models.Index(fields=['entry_id'],
                               name='corptools_c_entry_i_09f70e_idx'),
        ),
        migrations.AddIndex(
            model_name='corporationwalletjournalentry',
            index=models.Index(fields=['ref_type'],
                               name='corptools_c_ref_typ_326ddd_idx'),
        ),
    ]



# Generated by Django 2.2.9 on 2020-05-07 11:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mapregion',
            name='description',
            field=models.CharField(default=None, max_length=255, null=True),
        ),
    ]



import json
import xml.etree.ElementTree as ET
from datetime import timedelta

from django_celery_beat.models import CrontabSchedule, PeriodicTask

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.shortcuts import redirect, render
from django.template import TemplateDoesNotExist
from django.urls import reverse
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import _check_callback, token_required
from esi.views import sso_redirect

from . import __version__, app_settings
from .api.corporation import dashboards
from .forms import UploadForm
from .models import (
    CharacterAudit, CorpAsset, CorporationAudit, CorptoolsConfiguration,
    EveItemCategory, EveItemDogmaAttribute, EveItemGroup, EveItemType,
    EveLocation, EveName, InvTypeMaterials, MapConstellation, MapJumpBridge,
    MapRegion, MapSystem, MapSystemGate, SkillList, Structure,
)
from .tasks import (
    check_account, clear_all_etags, update_all_characters, update_all_corps,
    update_all_eve_names, update_all_locations, update_character, update_corp,
    update_or_create_map,
)

logger = get_extension_logger(__name__)

CORP_REQUIRED_SCOPES = [

    # Tracking
    'esi-corporations.track_members.v1',
    'esi-corporations.read_titles.v1',
    'esi-corporations.read_corporation_membership.v1',
    'esi-killmails.read_corporation_killmails.v1',

    # Moons
    'esi-industry.read_corporation_mining.v1',

    # Structures
    'esi-planets.read_customs_offices.v1',
    'esi-corporations.read_starbases.v1',
    'esi-corporations.read_structures.v1',

    # Wallets
    'esi-wallet.read_corporation_wallets.v1',
    'esi-markets.read_corporation_orders.v1',
    'esi-industry.read_corporation_jobs.v1',
    'esi-corporations.read_divisions.v1',

    # Assets
    'esi-assets.read_corporation_assets.v1',

    # All...
    'esi-search.search_structures.v1',
    'esi-universe.read_structures.v1',
    'esi-characters.read_corporation_roles.v1',

]


@login_required
@token_required(scopes=app_settings.get_character_scopes())
def add_char(request, token):
    CharacterAudit.objects.update_or_create(
        character=EveCharacter.objects.get_character_by_id(token.character_id))
    update_character.apply_async(args=[token.character_id], kwargs={
                                 "force_refresh": True}, priority=6)
    return redirect('corptools:reactmain', character_id=token.character_id)


@login_required
@token_required(scopes=CORP_REQUIRED_SCOPES)
def add_corp(request, token):
    char = EveCharacter.objects.get_character_by_id(token.character_id)
    corp, created = EveCorporationInfo.objects.get_or_create(corporation_id=char.corporation_id,
                                                             defaults={'member_count': 0,
                                                                       'corporation_ticker': char.corporation_ticker,
                                                                       'corporation_name': char.corporation_name
                                                                       })
    CorporationAudit.objects.update_or_create(corporation=corp)
    update_all_corps.apply_async(priority=6)
    return redirect('corptools:corp_react')


@login_required
def add_corp_section(request, *args, **kwargs):

    tracking = request.GET.get('t', False)
    assets = request.GET.get('a', False)
    structures = request.GET.get('s', False)
    wallets = request.GET.get('w', False)
    moons = request.GET.get('m', False)
    pocos = request.GET.get('p', False)
    contracts = request.GET.get('c', False)

    # if we're coming back from SSO with a new token, return it
    token = _check_callback(request)
    if token:
        logger.debug(
            "Got new token from %s session %s. Returning to view.",
            request.user,
            request.session.session_key[:5]
        )
        char = EveCharacter.objects.get_character_by_id(token.character_id)
        corp, created = EveCorporationInfo.objects.get_or_create(corporation_id=char.corporation_id,
                                                                 defaults={'member_count': 0,
                                                                           'corporation_ticker': char.corporation_ticker,
                                                                           'corporation_name': char.corporation_name
                                                                           })
        CorporationAudit.objects.update_or_create(corporation=corp)
        update_corp.apply_async(priority=6, args=[char.corporation_id])

        return redirect("{}#status".format(reverse('corptools:corp_react')))

    scopes = app_settings._corp_scopes_base

    if tracking:
        scopes += app_settings._corp_scopes_tracking

    if structures:
        scopes += app_settings._corp_scopes_structures

    if moons:
        scopes += app_settings._corp_scopes_moons

    if wallets:
        scopes += app_settings._corp_scopes_wallets

    if assets and not pocos:
        scopes += app_settings._corp_scopes_assets

    if pocos:
        scopes += app_settings._corp_scopes_pocos
        scopes += app_settings._corp_scopes_assets

    if contracts:
        scopes += app_settings._corp_scopes_contracts
    # user has selected to add a new token
    return sso_redirect(request, scopes=scopes)


@login_required
@permission_required('corptools.view_characteraudit')
def corptools_menu(request):
    # get available models
    cas = CharacterAudit.objects.visible_to(request.user)\
        .select_related('character__character_ownership__user__profile__main_character',
                        'character__characteraudit')\
        .prefetch_related('character__character_ownership__user__character_ownerships')\
        .prefetch_related('character__character_ownership__user__character_ownerships__character')\
        .prefetch_related('character__character_ownership__user__character_ownerships__character__characteraudit')\


    chars = {}
    orphans = []
    for char in cas:
        try:
            main = char.character.character_ownership.user.profile.main_character
            if main:
                if main.character_name not in chars:
                    chars[str(main.character_id)] = {'main': main,
                                                     'audit': char}
            else:
                orphans.append(char)
        except ObjectDoesNotExist:
            orphans.append(char)

    if len(chars) == 1:
        return redirect('corptools:overview', chars[list(chars.keys())[0]]['main'].character_id)

    return render(request, 'corptools/menu.html', context={'characters': chars, 'orphans': orphans})


@login_required
@permission_required('corptools.admin')
def admin(request):
    # get available models
    names = EveName.objects.all().count()
    types = EveItemType.objects.all().count()
    dogma = EveItemDogmaAttribute.objects.all().count()
    groups = EveItemGroup.objects.all().count()
    categorys = EveItemCategory.objects.all().count()
    type_mets = InvTypeMaterials.objects.count()
    regions = MapRegion.objects.all().count()
    constellations = MapConstellation.objects.all().count()
    systems = MapSystem.objects.all().count()
    gates = MapSystemGate.objects.all().count()
    location = EveLocation.objects.all().count()
    bridges = MapJumpBridge.objects.all().count()

    characters = CharacterAudit.objects.all().count()
    actives = CharacterAudit.get_oldest_qs().count()
    skilllists = SkillList.objects.all().count()
    corpations = CorporationAudit.objects.all().count()

    char_tasks = PeriodicTask.objects.filter(
        task='corptools.tasks.update_subset_of_characters', enabled=True).count()
    corp_tasks = PeriodicTask.objects.filter(
        task='corptools.tasks.update_all_corps', enabled=True).count()

    context = {
        "names": names,
        "types": types,
        "dogma": dogma,
        "groups": groups,
        "categorys": categorys,
        "characters": characters,
        "active_chars": actives,
        "skilllists": skilllists,
        "corpations": corpations,
        "type_mets": type_mets,
        "regions": regions,
        "constellations": constellations,
        "systems": systems,
        "location": location,
        "bridges": bridges,
        "gates": gates,
        "char_tasks": char_tasks,
        "corp_tasks": corp_tasks,
        "form": UploadForm(),
        "ct_config": CorptoolsConfiguration.get_solo(),
        "app_settings": app_settings
    }

    return render(request, 'corptools/admin.html', context=context)


@login_required
@permission_required('corptools.admin')
def admin_run_tasks(request):
    if request.method == 'POST':
        if request.POST.get('run_universe'):
            messages.info(request, "Queued update_or_create_map")
            update_or_create_map.apply_async(priority=6)
        if request.POST.get('run_update_all'):
            messages.info(request, "Queued update_all_characters")
            update_all_characters.apply_async(priority=6)
        if request.POST.get('run_update_eve_models'):
            messages.info(request, "Queued update_all_eve_names")
            update_all_eve_names.apply_async(priority=6)
        if request.POST.get('run_corp_updates'):
            messages.info(request, "Queued update_all_corps")
            update_all_corps.apply_async(priority=6)
        if request.POST.get('run_locations'):
            messages.info(request, "Queued update_all_locations")
            update_all_locations.apply_async(priority=6)
        if request.POST.get('run_clear_etag'):
            messages.info(request, "Queued clear_all_etags")
            clear_all_etags.apply_async(priority=1)

    return redirect('corptools:admin')


@login_required
@permission_required('corptools.admin')
def admin_create_tasks(request):
    schedule_char, _ = CrontabSchedule.objects.get_or_create(minute='15,45',
                                                             hour='*',
                                                             day_of_week='*',
                                                             day_of_month='*',
                                                             month_of_year='*',
                                                             timezone='UTC'
                                                             )

    schedule_corp, _ = CrontabSchedule.objects.get_or_create(minute='30',
                                                             hour='*',
                                                             day_of_week='*',
                                                             day_of_month='*',
                                                             month_of_year='*',
                                                             timezone='UTC'
                                                             )

    PeriodicTask.objects.update_or_create(
        task='corptools.tasks.update_subset_of_characters',
        defaults={
            'crontab': schedule_char,
            'name': 'Character Audit Rolling Update',
            'enabled': True
        }
    )
    # if created:
    #    messages.info(request, "Created Character Task")
    # else:
    #    messages.info(request, "Reset Character Task to defaults")

    PeriodicTask.objects.update_or_create(
        task='corptools.tasks.update_all_corps',
        defaults={
            'crontab': schedule_corp,
            'name': 'Corporation Audit Update',
            'enabled': True
        }
    )
    # if created:
    #    messages.info(request, "Created Corporation Task")
    # else:
    #    messages.info(request, "Reset Corporation Task to defaults")

    # https://github.com/celery/django-celery-beat/issues/106
    messages.info(
        request, "Created/Reset Character and Corporation Task to defaults")

    return redirect('corptools:admin')


@login_required
@permission_required('corptools.admin')
def admin_add_pyfa_xml(request):
    if request.method == 'POST':
        form = UploadForm(request.POST, request.FILES)
        if form.is_valid():
            xml_file = form.files.get('file')
            if xml_file.content_type == 'text/xml':
                skills = {}
                tree = ET.parse(xml_file)
                root = tree.getroot()
                if root.tag == "plan":
                    for child in root:
                        if child.tag == "entry":
                            skill = child.attrib.get('skill')
                            level = child.attrib.get('level')
                            if skill in skills:
                                if level > skills[skill]:
                                    skills[skill] = level
                            else:
                                skills[skill] = level
                sl, created = SkillList.objects.update_or_create(name=request.POST["name"],
                                                                 defaults={
                    "skill_list": json.dumps(skills)
                })

                messages.success(request, "File Uploaded and Processed! {}: {}".format(
                    ("Created" if created else "Updated"),
                    request.POST["name"]
                ))
            else:
                messages.error(
                    request, "File Upload Failed! Must be XML Exported from PYFA!")
        else:
            messages.error(request, "File Upload Failed! Invalid Form")
    else:
        messages.error(
            request, "File Upload Failed! What are you doing your not meant to be here?")

    return redirect('corptools:admin')


@login_required
@permission_required('corptools.admin')
def skill_list_editor(request):
    pass


@login_required
def update_account(request, character_id):
    check_account.apply_async(args=[character_id], priority=6)
    messages.success(request, "Requested an update task.")
    return redirect('corptools:view')


@login_required
def react_menu(request):
    return redirect('corptools:reactlegacymain', request.user.profile.main_character.character_id)


@login_required
def react_menu_v4(request):
    if request.path.startswith('/audit/r_beta/corp/glance'):
        return render(request, 'corptools/base_bs5.html', context={"version": __version__, "app_name": "corptools/bs5", "page_title": "Audit"})
    return redirect('corptools:reactmain', request.user.profile.main_character.character_id)


@login_required
def react_corp_beta(request):
    return render(request, 'corptools/base_bs5.html', context={"version": __version__, "app_name": "corptools/bs5", "page_title": "Audit"})


@login_required
def react_main(request, character_id):
    try:
        return render(request, 'corptools/base_bs5.html', context={"version": __version__, "app_name": "corptools/bs5", "page_title": "Audit"})
    except TemplateDoesNotExist:
        pass

    return render(request, 'corptools/character/react_base.html', context={"version": __version__, "app_name": "corptools/char", "page_title": "Character Audit"})


@login_required
def v3_ui_render(request, character_id):
    return render(request, 'corptools/character/react_base.html', context={"version": __version__, "app_name": "corptools/char", "page_title": "Character Audit"})


@login_required
def react_corp(request):
    return render(request, 'corptools/corporation/react_base.html', context={"version": __version__, "app_name": "corptools/corp", "page_title": "Corporation Audit"})


@login_required
def fuel_levels(request):
    if not (request.user.has_perm('corptools.own_corp_manager')
            or request.user.has_perm('corptools.alliance_corp_manager')
            or request.user.has_perm('corptools.state_corp_manager')
            or request.user.has_perm('corptools.global_corp_manager')
            or request.user.has_perm('corptools.holding_corp_structures')):
        raise PermissionDenied("No perms to view")

    # hourly fuel reqs [ cit, eng, ref, flex ]
    citadel_service_mods = {
        'Clone Bay': [7.5, 10, 10, 10],
        'Market': [30, 40, 40, 40],
        'Manufacturing (Capitals)': [24, 18, 24, 24],
        # how to detect this
        'Standup Hyasyoda Research Lab': [10, 7.5, 10, 10],
        'Invention': [12, 9, 12, 12],
        'Manufacturing (Standard)': [12, 9, 12, 12],
        'Blueprint Copying': [12, 9, 12, 12],
        'Material Efficiency Research': [0, 0, 0, 0],  # part of above
        'Time Efficiency Research': [0, 0, 0, 0],  # part of above
        'Manufacturing (Super Capitals)': [36, 27, 36, 36],
        'Composite Reactions': [15, 15, 12, 15],
        'Hybrid Reactions': [15, 15, 12, 15],
        'Moon Drilling': [5, 5, 5, 5],
        'Biochemical Reactions': [15, 15, 12, 15],
        'Reprocessing': [10, 10, 7.5, 10],
        'Jump Access': [0, 0, 0, 15],  # large to show errors
        'Cynosural Jammer': [0, 0, 0, 40],
        'Jump Gate Access': [0, 0, 0, 30],
        'Automatic Moon Drilling': [0, 0, 0, 5]
    }

    cit = [35833, 47516, 47512, 47513, 47514, 47515, 35832, 35834]
    eng = [35827, 35826, 35825]
    ref = [35835, 35836]
    fle = [37534, 35841, 35840, 81826]

    flex_fuel_types = [81143]

    all_structures = Structure.get_visible(request.user).select_related(
        'corporation', 'corporation__corporation', 'system_name', 'type_name',
        'system_name__constellation', 'system_name__constellation__region'
    ).prefetch_related('structureservice_set')

    structure_tree = []
    total_hourly_fuel = 0
    for s in all_structures:
        structure_hourly_fuel = 0
        structure_type = 99

        if s.type_id in cit:
            structure_type = 0
        elif s.type_id in eng:
            structure_type = 1
        elif s.type_id in ref:
            structure_type = 2
        elif s.type_id in fle:
            structure_type = 3

        for service in s.structureservice_set.all():
            if service.state == 'online':
                fuel_use = citadel_service_mods[service.name][structure_type]
                total_hourly_fuel += fuel_use
                structure_hourly_fuel += fuel_use

        hours = 0

        if s.fuel_expires is not None:
            hours = (s.fuel_expires - timezone.now()).total_seconds() // 3600

        extras = None

        if s.type_id == 81826:
            fuels = CorpAsset.objects.filter(
                type_id__in=flex_fuel_types,
                location_id=s.structure_id
            ).select_related(
                'type_name'
            )
            out = {
                "qty": 0,
            }
            for itm in fuels:
                out["name"] = itm.type_name.name
                out["qty"] += itm.quantity

            out["expires"] = timezone.now() + timedelta(hours=out["qty"] / 88)
            if out['qty'] > 0:
                extras = out

        structure_tree.append(
            {
                'structure': s,
                'fuel_req': structure_hourly_fuel,
                "current_blocks": int(hours * structure_hourly_fuel),
                "extra_fuel_info": extras,
                "90day": max(
                    int(
                        (
                            structure_hourly_fuel * 90 * 24
                        ) - (
                            hours * structure_hourly_fuel
                        )
                    ),
                    0
                )
            }
        )

    context = {
        'structures': structure_tree,
        'total_hourly_fuel': total_hourly_fuel,
    }
    return render(request, 'corptools/dashboards/fuel_dash.html', context=context)


@login_required
def metenox_levels(request):
    if not (request.user.has_perm('corptools.own_corp_manager')
            or request.user.has_perm('corptools.alliance_corp_manager')
            or request.user.has_perm('corptools.state_corp_manager')
            or request.user.has_perm('corptools.global_corp_manager')
            or request.user.has_perm('corptools.holding_corp_structures')):
        raise PermissionDenied("No perms to view")

    context = {
        'structures': dashboards.DashboardApiEndpoints.get_dashboard_drills_levels(request),
    }
    return render(request, 'corptools/dashboards/metenox_dash.html', context=context)



from typing import List

from ninja import NinjaAPI

from django.db.models import F
from django.utils.translation import gettext as _

from allianceauth.services.hooks import get_extension_logger

from corptools import models
from corptools.api import schema

logger = get_extension_logger(__name__)


class SearchApiEndpoints:

    tags = ["Search"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "/search/system/{search_text}",
            response={200: List[schema.EveName]},
            tags=["Search"]
        )
        def get_system_search(request, search_text: str, limit: int = 10):
            if not request.user.is_superuser:
                return 403, _("Hard no pall!")
            return models.MapSystem.objects.filter(name__icontains=search_text).values("name", id=F("system_id"))[:limit]

        @api.get(
            "/search/location/{search_text}",
            response={200: List[schema.EveName]},
            tags=["Search"]
        )
        def get_location_search(request, search_text: str, limit: int = 10):
            if not request.user.is_superuser:
                return 403, _("Hard no pall!")

            return models.EveLocation.objects.filter(location_name__icontains=search_text).exclude(location_id__lte=0).values(name=F("location_name"), id=F("location_id"))[:limit]

        @api.get(
            "/search/item/group/{search_text}",
            response={200: List[schema.EveName]},
            tags=["Search"]
        )
        def get_group_search(request, search_text: str, limit: int = 10):
            if not request.user.is_superuser:
                return 403, _("Hard no pall!")

            return models.EveItemGroup.objects.filter(name__icontains=search_text).values("name", id=F("group_id"))[:limit]



# Generated by Django 4.2.4 on 2023-08-20 00:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0090_rename_cache_expire_assets_corporationaudit_last_change_assets_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='assetsfilter',
            name='count_message_only',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='fullyloadedfilter',
            name='count_message_only',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='skillfilter',
            name='count_message_only',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.12 on 2020-09-15 11:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0032_auto_20200901_0537'),
    ]

    operations = [
        migrations.AddField(
            model_name='skilllist',
            name='order_weight',
            field=models.IntegerField(default=0),
        ),
    ]



# Generated by Django 2.2.9 on 2020-05-11 10:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0010_auto_20200511_0853'),
    ]

    operations = [
        migrations.CreateModel(
            name='CorporationWalletDivision',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('balance', models.DecimalField(decimal_places=2, max_digits=20)),
                ('division', models.IntegerField()),
                ('corporation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                  related_name='corporation_division', to='corptools.CorporationAudit')),
            ],
        ),
        migrations.CreateModel(
            name='CorporationWalletJournalEntry',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', models.DecimalField(decimal_places=2,
                                               default=None, max_digits=20, null=True)),
                ('balance', models.DecimalField(decimal_places=2,
                                                default=None, max_digits=20, null=True)),
                ('context_id', models.BigIntegerField(default=None, null=True)),
                ('context_id_type', models.CharField(choices=[('structure_id', 'structure_id'), ('station_id', 'station_id'), ('market_transaction_id', 'market_transaction_id'), ('character_id', 'character_id'), ('corporation_id', 'corporation_id'), (
                    'alliance_id', 'alliance_id'), ('eve_system', 'eve_system'), ('industry_job_id', 'industry_job_id'), ('contract_id', 'contract_id'), ('planet_id', 'planet_id'), ('system_id', 'system_id'), ('type_id', 'type_id')], default=None, max_length=30, null=True)),
                ('date', models.DateTimeField()),
                ('description', models.CharField(max_length=500)),
                ('first_party_id', models.IntegerField(default=None, null=True)),
                ('entry_id', models.BigIntegerField()),
                ('reason', models.CharField(default=None, max_length=500, null=True)),
                ('ref_type', models.CharField(max_length=72)),
                ('second_party_id', models.IntegerField(default=None, null=True)),
                ('tax', models.DecimalField(decimal_places=2,
                                            default=None, max_digits=20, null=True)),
                ('tax_receiver_id', models.IntegerField(default=None, null=True)),
                ('division', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.CorporationWalletDivision')),
                ('first_party_name', models.ForeignKey(default=None, null=True,
                                                       on_delete=django.db.models.deletion.SET_NULL, related_name='corp_first_party', to='corptools.EveName')),
                ('second_party_name', models.ForeignKey(default=None, null=True,
                                                        on_delete=django.db.models.deletion.SET_NULL, related_name='corp_second_party', to='corptools.EveName')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CharacterWalletJournalEntry',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', models.DecimalField(decimal_places=2,
                                               default=None, max_digits=20, null=True)),
                ('balance', models.DecimalField(decimal_places=2,
                                                default=None, max_digits=20, null=True)),
                ('context_id', models.BigIntegerField(default=None, null=True)),
                ('context_id_type', models.CharField(choices=[('structure_id', 'structure_id'), ('station_id', 'station_id'), ('market_transaction_id', 'market_transaction_id'), ('character_id', 'character_id'), ('corporation_id', 'corporation_id'), (
                    'alliance_id', 'alliance_id'), ('eve_system', 'eve_system'), ('industry_job_id', 'industry_job_id'), ('contract_id', 'contract_id'), ('planet_id', 'planet_id'), ('system_id', 'system_id'), ('type_id', 'type_id')], default=None, max_length=30, null=True)),
                ('date', models.DateTimeField()),
                ('description', models.CharField(max_length=500)),
                ('first_party_id', models.IntegerField(default=None, null=True)),
                ('entry_id', models.BigIntegerField()),
                ('reason', models.CharField(default=None, max_length=500, null=True)),
                ('ref_type', models.CharField(max_length=72)),
                ('second_party_id', models.IntegerField(default=None, null=True)),
                ('tax', models.DecimalField(decimal_places=2,
                                            default=None, max_digits=20, null=True)),
                ('tax_receiver_id', models.IntegerField(default=None, null=True)),
                ('character', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.CharacterAudit')),
                ('first_party_name', models.ForeignKey(default=None, null=True,
                                                       on_delete=django.db.models.deletion.SET_NULL, related_name='char_first_party', to='corptools.EveName')),
                ('second_party_name', models.ForeignKey(default=None, null=True,
                                                        on_delete=django.db.models.deletion.SET_NULL, related_name='char_second_party', to='corptools.EveName')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from typing import List

from ninja import NinjaAPI

from django.utils.translation import gettext as _

from corptools import models, providers
from corptools.api import schema
from corptools.api.helpers import get_alts_queryset, get_main_character


class SkillApiEndpoints:

    tags = ["Account"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "account/{character_id}/skills",
            response={200: List[schema.CharacterSkills], 403: str},
            tags=self.tags
        )
        def get_character_skills(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            skills = models.Skill.objects.filter(character__character__in=characters)\
                .select_related('character__character', 'skill_name', "skill_name__group")

            totals = models.SkillTotals.objects.filter(character__character__in=characters)\
                .select_related('character__character')

            output = {}

            for s in skills:
                if s.character_id not in output:
                    output[s.character_id] = {
                        "character": s.character.character,
                        "skills": [],
                        "total_sp": 0,
                        "unallocated_sp": 0
                    }
                output[s.character_id]["skills"].append(
                    {
                        "group": s.skill_name.group.name,
                        "skill": s.skill_name.name,
                        "sp": s.skillpoints_in_skill,
                        "level": s.trained_skill_level,
                        "active": s.active_skill_level,
                    }
                )

            for t in totals:
                if t.character_id in output:
                    output[t.character_id]["unallocated_sp"] = t.unallocated_sp
                    output[t.character_id]["total_sp"] = t.total_sp

            return list(output.values())

        @api.get(
            "account/{character_id}/skill/history",
            response={200: List[schema.CharacterSkillHistory], 403: str},
            tags=self.tags
        )
        def get_character_skill_history(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            skill_history = models.SkillTotalHistory.objects.filter(character__character__in=characters)\
                .select_related('character__character').order_by("date")

            output = {}

            for s in skill_history:
                if s.character_id not in output:
                    output[s.character_id] = {
                        "character": s.character.character,
                        "history": [],
                    }
                output[s.character_id]["history"].append(
                    {
                        "date": s.date,
                        "sp": s.sp,
                        "total_sp": s.total_sp,
                        "unallocated_sp": s.unallocated_sp,
                    }
                )

            return list(output.values())

        @api.get(
            "account/{character_id}/skillqueues",
            response={200: List[schema.CharacterQueue], 403: str},
            tags=self.tags
        )
        def get_character_skillqueues(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            skills = models.SkillQueue.objects.filter(character__character__in=characters)\
                .select_related('character__character', 'skill_name', "skill_name__group")

            skill_totals = models.Skill.objects.filter(character__character__in=characters)\
                .select_related('character__character', 'skill_name', "skill_name__group")

            skl_ttl = {}

            for s in skill_totals:
                if s.character_id not in skl_ttl:
                    skl_ttl[s.character_id] = {}
                skl_ttl[s.character_id][s.skill_name.name] = s.trained_skill_level

            output = {}
            for c in characters:
                output[c.character_id] = {
                    "character": c,
                    "queue": [],
                }

            for s in skills:
                if s.character.character.character_id not in output:
                    output[s.character.character.character_id] = {
                        "character": s.character.character,
                        "queue": [],
                    }
                output[s.character.character.character_id]["queue"].append(
                    {
                        "position": s.queue_position,
                        "group": s.skill_name.group.name,
                        "skill": s.skill_name.name,
                        "end_level": s.finish_level,
                        "start_sp": s.level_start_sp,
                        "end_sp": s.level_end_sp,
                        "start": s.start_date,
                        "end": s.finish_date,
                        "current_level": skl_ttl.get(s.character.character_id, {}).get(s.skill_name.name, 0)
                    }
                )

            return list(output.values())

        @api.get(
            "account/{character_id}/doctrines",
            response={200: List[schema.CharacterDoctrines], 403: str},
            tags=self.tags
        )
        def get_character_doctrines(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            skilllists = providers.skills.get_and_cache_user(
                main.character_ownership.user_id)
            visibles = list(models.SkillList.objects.filter(
                show_on_audit=True).values_list("name", flat=True))
            output = {}
            for c in characters:
                output[c.character_id] = {
                    "character": c,
                    "doctrines": {},
                    "skills": {},
                }

            for k, s in skilllists['skills_list'].items():
                for k, d in s["doctrines"].items():
                    # filter out hidden items
                    if k in visibles:
                        output[s['character_id']]["doctrines"][k] = d
                output[s['character_id']]["skills"] = s["skills"]

            return list(output.values())



# Generated by Django 3.1.1 on 2020-09-28 00:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0033_skilllist_order_weight'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='characterasset',
            index=models.Index(fields=['location_id'],
                               name='corptools_c_locatio_b58222_idx'),
        ),
        migrations.AddIndex(
            model_name='characterasset',
            index=models.Index(fields=['type_id'],
                               name='corptools_c_type_id_05ccc2_idx'),
        ),
        migrations.AddIndex(
            model_name='characterasset',
            index=models.Index(fields=['item_id'],
                               name='corptools_c_item_id_1800a8_idx'),
        ),
        migrations.AddIndex(
            model_name='corpasset',
            index=models.Index(fields=['location_id'],
                               name='corptools_c_locatio_a86faf_idx'),
        ),
        migrations.AddIndex(
            model_name='corpasset',
            index=models.Index(fields=['type_id'],
                               name='corptools_c_type_id_e9d79e_idx'),
        ),
        migrations.AddIndex(
            model_name='corpasset',
            index=models.Index(fields=['item_id'],
                               name='corptools_c_item_id_41066b_idx'),
        ),
    ]



# Generated by Django 3.2.8 on 2022-01-08 06:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0059_invtypematerials_met_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='BridgeOzoneLevel',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('station_id', models.CharField(max_length=500)),
                ('quantity', models.BigIntegerField()),
                ('used', models.BigIntegerField(default=0)),
                ('date', models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



import re
from typing import Union

import requests
from celery import chain, shared_task

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db.models import F, Q
from django.db.models.aggregates import Sum
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from esi.errors import TokenError
from esi.models import Token

from corptools.task_helpers.etag_helpers import NotModifiedError, etag_results
from corptools.task_helpers.update_tasks import fetch_location_name

from .. import providers
from ..models import (
    BridgeOzoneLevel, CharacterAudit, CorpAsset, CorporateContract,
    CorporateContractItem, CorporationAudit, CorporationWalletDivision,
    CorporationWalletJournalEntry, CorptoolsConfiguration, EveItemType,
    EveLocation, EveName, MapJumpBridge, MapSystem, MapSystemPlanet, Poco,
    Structure, StructureCelestial, StructureService,
)
from . import sanitize_location_flag

logger = get_extension_logger(__name__)


def get_corp_token(corp_id: int, scopes: list, req_roles: Union[list, None, bool]):
    """
    Helper method to get a token for a specific character from a specific corp with specific scopes, where
    a character has specific in game roles.
    :param corp_id: Corp to filter on.
    :param scopes: array of ESI scope strings required on the token
    :param req_roles: array of roles, one of which is required on the character.
    :return: :class:esi.models.Token or None
    """

    # always add roles scope as a requirement.
    if 'esi-characters.read_corporation_roles.v1' not in scopes:
        scopes.append("esi-characters.read_corporation_roles.v1")

    # Find all characters in the corporation known to auth.
    char_ids = EveCharacter.objects.filter(
        corporation_id=corp_id).values('character_id')

    # find all tokens for the corp, with the scopes.
    tokens = Token.objects.filter(
        character_id__in=char_ids
    ).require_scopes(scopes)

    # loop to check the roles and break on first correct match
    for token in tokens:
        try:
            if req_roles:  # There are endpoints with no requirements
                roles = providers.esi.client.Character.get_characters_character_id_roles(
                    character_id=token.character_id,
                    token=token.valid_access_token()
                ).result()

                has_roles = False

                # do we have the roles.
                for role in roles.get('roles', []):
                    if role in req_roles:
                        has_roles = True
                        break

                if has_roles:
                    return token  # return the token
                else:
                    pass  # next! TODO should we flag this character?
            else:
                return token  # no roles check needed return the token
        except TokenError as e:
            #  I've had invalid tokens in auth that refresh but don't actually work
            logger.error(f"Token Error ID: {token.pk} ({e})")

    return None


# pagnated results
def update_corp_wallet_journal(corp_id, wallet_division, full_update=False):
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)

    division = CorporationWalletDivision.objects.get(
        corporation=audit_corp, division=wallet_division)

    logger.debug("Updating wallet transactions for: {} (Div: {})".format(
        audit_corp.corporation.corporation_name, division))

    req_scopes = ['esi-wallet.read_corporation_wallets.v1',
                  'esi-characters.read_corporation_roles.v1']

    req_roles = ['CEO', 'Director', 'Accountant', 'Junior_Accountant']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"

    _current_journal = set(list(CorporationWalletJournalEntry.objects.filter(
        division=division).order_by('-date').values_list('entry_id', flat=True)[:20000]))
    _current_eve_ids = set(list(
        EveName.objects.all().values_list('eve_id', flat=True)))

    current_page = 1
    total_pages = 1
    _new_names = []
    while current_page <= total_pages:
        journal_items = providers.esi.client.Wallet.get_corporations_corporation_id_wallets_division_journal(corporation_id=audit_corp.corporation.corporation_id,
                                                                                                             division=wallet_division,
                                                                                                             page=current_page,
                                                                                                             token=token.valid_access_token())
        journal_items.request_config.also_return_response = True
        journal_items, headers = journal_items.result()
        total_pages = int(headers.headers['X-Pages'])
        logger.debug(
            f"CT: Corp {corp_id} Div {wallet_division}, Page:{current_page}/{total_pages} ({len(journal_items)})")
        _min_time = timezone.now()
        items = []
        for item in journal_items:
            if _min_time > item.get('date'):
                _min_time = item.get('date')

            if item.get('id') not in _current_journal:
                if item.get('second_party_id') not in _current_eve_ids:
                    _new_names.append(item.get('second_party_id'))
                    _current_eve_ids.add(item.get('second_party_id'))
                if item.get('first_party_id') not in _current_eve_ids:
                    _new_names.append(item.get('first_party_id'))
                    _current_eve_ids.add(item.get('first_party_id'))
                wallet_item = CorporationWalletJournalEntry(division=division,
                                                            amount=item.get(
                                                                'amount'),
                                                            balance=item.get(
                                                                'balance'),
                                                            context_id=item.get(
                                                                'context_id'),
                                                            context_id_type=item.get(
                                                                'context_id_type'),
                                                            date=item.get(
                                                                'date'),
                                                            description=item.get(
                                                                'description'),
                                                            first_party_id=item.get(
                                                                'first_party_id'),
                                                            first_party_name_id=item.get(
                                                                'first_party_id'),
                                                            entry_id=item.get(
                                                                'id'),
                                                            reason=item.get(
                                                                'reason'),
                                                            ref_type=item.get(
                                                                'ref_type'),
                                                            second_party_id=item.get(
                                                                'second_party_id'),
                                                            second_party_name_id=item.get(
                                                                'second_party_id'),
                                                            tax=item.get(
                                                                'tax'),
                                                            tax_receiver_id=item.get(
                                                                'tax_receiver_id'),
                                                            )
                items.append(wallet_item)
            # else:
                # short cct
                # if not full_update:
                # total_pages = 0
                # break
        logger.info(
            f"CT: Corp {corp_id} Div {wallet_division}, Page: {current_page}, New Transactions! {len(items)}, New Names {_new_names}")
        created_names = EveName.objects.create_bulk_from_esi(_new_names)

        if created_names:
            CorporationWalletJournalEntry.objects.bulk_create(items)
        else:
            raise Exception("DB Fail")

        current_page += 1

    logger.info(
        f"CT: Corp {corp_id} Div {wallet_division}, OLDEST DATA! {audit_corp} {_min_time}")


def update_corporation_transactions(corp_id, wallet_division, full_update=False):
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)

    division = CorporationWalletDivision.objects.get(
        corporation=audit_corp, division=wallet_division)

    logger.debug("Updating market transactions for: {} (Div: {})".format(
        audit_corp.corporation.corporation_name, division))

    req_scopes = ['esi-wallet.read_corporation_wallets.v1',
                  'esi-characters.read_corporation_roles.v1']

    req_roles = ['CEO', 'Director', 'Accountant', 'Junior_Accountant']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"

    try:
        journal_items_ob = providers.esi.client.Wallet.get_corporations_corporation_id_wallets_division_transactions(
            corporation_id=audit_corp.corporation.corporation_id,
            division=wallet_division)

        journal_items = etag_results(
            journal_items_ob, token, force_refresh=full_update)

        _current_journal = CorporationWalletJournalEntry.objects.filter(
            character=audit_corp,
            context_id_type="market_transaction_id",
            # Max items from ESI
            reason__exact="").values_list('context_id', flat=True)[:2500]

        # _new_names = []

        # items = []
        for item in journal_items:
            if item.get('transaction_id') in _current_journal:
                type_name, _ = EveItemType.objects.get_or_create_from_esi(
                    item.get('type_id'))

    except NotModifiedError:
        logger.info(
            f"CT: No New market transaction data for: {audit_corp.corporation.corporation_name}")
        pass

    return f"CT: Finished market transactions for: {audit_corp.corporation.corporation_name}"


def update_corporation_pocos(corp_id, full_update=False):
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)

    logger.debug("Updating Pocos for: {}".format(
        audit_corp.corporation.corporation_name))

    req_scopes = ['esi-planets.read_customs_offices.v1']

    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"

    try:
        poco_ob = providers.esi.client.Planetary_Interaction.get_corporations_corporation_id_customs_offices(
            corporation_id=audit_corp.corporation.corporation_id)

        poco_data = etag_results(
            poco_ob, token, force_refresh=full_update)

        # Get all Poco Names ( planet name here )
        _all_ids = [p.get("office_id") for p in poco_data]

        token_assets = get_corp_token(
            corp_id, ['esi-assets.read_corporation_assets.v1'], req_roles)

        _all_names = []

        for id_chunk in providers.esi.chunk_ids(_all_ids):
            _all_names += providers.esi.client.Assets.post_corporations_corporation_id_assets_names(
                corporation_id=corp_id,
                item_ids=id_chunk,
                token=token_assets.valid_access_token()
            ).result()

        _planets = []
        _office_to_names = {}
        _planet_to_office = {}
        for n in _all_names:
            pn = n['name'][16: -1]
            _office_to_names[n['item_id']] = {"n": n['name'], "p": None}
            _planet_to_office[pn] = n['item_id']
            _planets += pn

        # Ensure all planets are in DB
        _all_system_ids = [p.get("system_id") for p in poco_data]

        _pids = []
        for system_id in _all_system_ids:
            _s = providers.esi.client.Universe.get_universe_systems_system_id(
                system_id=system_id).results()
            _pids += [_p['planet_id'] for _p in _s.get('planets', [])]

        for _p in _pids:
            p, _ = MapSystemPlanet.objects.get_or_create_from_esi(_p)
            if p.name in _planet_to_office:
                _office_to_names[_planet_to_office[p.name]]["p"] = p.planet_id

        for poco in poco_data:
            Poco.objects.update_or_create(
                office_id=poco.get('office_id'),
                corporation=audit_corp,
                defaults={
                    "alliance_tax_rate": poco.get('alliance_tax_rate'),
                    "allow_access_with_standings": poco.get('allow_access_with_standings'),
                    "allow_alliance_access": poco.get('allow_alliance_access'),
                    "bad_standing_tax_rate": poco.get('bad_standing_tax_rate'),
                    "corporation_tax_rate": poco.get('corporation_tax_rate'),
                    "excellent_standing_tax_rate": poco.get('excellent_standing_tax_rate'),
                    "good_standing_tax_rate": poco.get('good_standing_tax_rate'),
                    "neutral_standing_tax_rate": poco.get('neutral_standing_tax_rate'),
                    "office_id": poco.get('office_id'),
                    "reinforce_exit_end": poco.get('reinforce_exit_end'),
                    "reinforce_exit_start": poco.get('reinforce_exit_start'),
                    "standing_level": poco.get('standing_level'),
                    "system_id": poco.get('system_id'),
                    "system_name_id": poco.get('system_id'),
                    "name": _office_to_names.get(poco.get('office_id'))['n'],
                    "planet_id": _office_to_names.get(poco.get('office_id'))['p'],
                    "terrible_standing_tax_rate": poco.get('terrible_standing_tax_rate')
                }
            )

        Poco.objects.filter(corporation=audit_corp).exclude(
            office_id__in=_all_ids).delete()

    except NotModifiedError:
        logger.info("CT: No New Poco data for: {}".format(
            audit_corp.corporation.corporation_name))
        pass

    return f"CT: Finished Pocos for: {audit_corp.corporation.corporation_name}"


def update_corp_wallet_division(corp_id, full_update=False):  # pagnated results
    # logger.debug("Started wallet divs for: %s" % str(character_id))
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)

    req_scopes = ['esi-wallet.read_corporation_wallets.v1',
                  'esi-characters.read_corporation_roles.v1',
                  'esi-corporations.read_divisions.v1']
    req_roles = ['CEO', 'Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)
    names = {}

    if token:
        division_names = providers.esi.client.Corporation.get_corporations_corporation_id_divisions(corporation_id=audit_corp.corporation.corporation_id,
                                                                                                    token=token.valid_access_token()).result()
        for division in division_names.get('wallet'):
            names[division.get('division')] = division.get('name')

    req_roles = ['CEO', 'Director', 'Accountant', 'Junior_Accountant']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"

    divisions = providers.esi.client.Wallet.get_corporations_corporation_id_wallets(corporation_id=audit_corp.corporation.corporation_id,
                                                                                    token=token.valid_access_token()).result()

    for division in divisions:
        _division_item, _created = CorporationWalletDivision.objects \
            .update_or_create(corporation=audit_corp, division=division.get('division'),
                              defaults={'balance': division.get('balance'),
                                        'name': names.get(division.get('division'), "Unknown")})

        if _division_item:
            update_corp_wallet_journal(corp_id, division.get('division'),
                                       full_update=full_update)  # inline not async

    audit_corp.last_update_wallet = timezone.now()
    audit_corp.save()

    return f"Finished wallet divs for: {audit_corp.corporation.corporation_name}"


def update_corp_structures(corp_id, force_refresh=False):  # pagnated results
    # logger.debug("Started structures for: %s" % (str(character_id)))

    def _structures_db_update(_corporation, _structure, _name):
        str_type, _ = EveItemType.objects.get_or_create_from_esi(
            _structure.get('type_id'))
        _structure_ob, _created = Structure.objects.update_or_create(
            structure_id=_structure.get('structure_id'),
            corporation=_corporation,
            defaults={
                'fuel_expires': _structure.get('fuel_expires', None),
                'next_reinforce_apply': _structure.get('next_reinforce_apply', None),
                'next_reinforce_weekday': _structure.get('next_reinforce_weekday', None),
                'profile_id': _structure.get('profile_id', None),
                'reinforce_hour': _structure.get('reinforce_hour', None),
                'reinforce_weekday': _structure.get('reinforce_weekday', None),
                'state': _structure.get('state', None),
                'state_timer_end': _structure.get('state_timer_end', None),
                'state_timer_start': _structure.get('state_timer_start', None),
                'system_id': _structure.get('system_id', None),
                'type_id': _structure.get('type_id', None),
                'unanchors_at': _structure.get('unanchors_at', None),
                'name': _name,
                'system_name_id': _structure.get('system_id', None),
                'type_name': str_type
            })

        if _structure_ob:
            # _asset = None
            _location = None
            celestial = StructureCelestial.objects.filter(
                structure_id=_structure.get('structure_id'))

            if not celestial.exists():
                try:
                    # _asset = CorpAsset.objects.get(
                    #     item_id=_structure.get('structure_id'), corp=_corporation)
                    _req_scopes = ['esi-assets.read_corporation_assets.v1']
                    _req_roles = ['Director']
                    _token = get_corp_token(
                        _corporation.corporation.corporation_id, _req_scopes, _req_roles)
                    if _token:
                        locations = providers.esi.client.Assets.post_corporations_corporation_id_assets_locations(
                            corporation_id=_corporation.corporation.corporation_id,
                            item_ids=[_structure.get('structure_id')],
                            token=token.valid_access_token()).result()

                        _location = locations[0]

                        url = "https://www.fuzzwork.co.uk/api/nearestCelestial.php?x=%s&y=%s&z=%s&solarsystemid=%s" \
                            % ((str(_location['position'].get('x'))),
                               (str(_location['position'].get('y'))),
                                (str(_location['position'].get('z'))),
                                (str(_structure.get('system_id')))
                               )

                        r = requests.get(url)
                        fuzz_result = r.json()

                        celestial = StructureCelestial.objects.create(
                            structure_id=_structure.get('structure_id'),
                            celestial_name=fuzz_result.get('itemName')
                        )
                except ObjectDoesNotExist:
                    celestial = None
                except Exception:
                    # logging.exception("Messsage")
                    celestial = None
            else:
                celestial = celestial[0]

            # TODO fix this all up.
            if isinstance(celestial, StructureCelestial):
                if celestial is not None:
                    _structure_ob.closest_celestial = celestial
                    _structure_ob.save()
            else:
                celestial = None

            if _structure.get('services'):
                db_services = StructureService.objects.filter(
                    structure=_structure_ob)
                current_services = []
                for service in _structure.get('services'):
                    db_service = db_services.filter(name=service['name'])
                    if db_service.exists():
                        if db_service.count() == 1:
                            if db_service[0].state == service['state']:
                                current_services.append(db_service[0].id)
                                pass
                            else:
                                db_service.update(state=service['state'])
                                current_services.append(db_service[0].id)
                        else:
                            StructureService.objects.filter(structure=_structure_ob,
                                                            name=service['name']).delete()
                            new_service = StructureService.objects.create(structure=_structure_ob,
                                                                          state=service['state'],
                                                                          name=service['name'])
                            current_services.append(new_service.id)

                    else:
                        new_service = StructureService.objects.create(structure=_structure_ob,
                                                                      state=service['state'],
                                                                      name=service['name'])
                        current_services.append(new_service.id)
                db_services.exclude(id__in=current_services).delete()

        return _structure_ob, _created

    req_scopes = ['esi-corporations.read_structures.v1', 'esi-universe.read_structures.v1',
                  'esi-characters.read_corporation_roles.v1']

    req_roles = ['Director', 'Station_Manager']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"

    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)

    structure_ids = []
    operation = providers.esi.client.Corporation.get_corporations_corporation_id_structures(
        corporation_id=_corporation.corporation.corporation_id)
    try:
        structures = etag_results(
            operation, token, force_refresh=force_refresh)
    except NotModifiedError:
        _corporation.last_update_structures = timezone.now()
        _corporation.save()
        return f"No New structure data for: {_corporation}"

    for structure in structures:
        name = str(structure.get('structure_id'))
        if structure.get('name'):
            name = structure.get('name')
            EveLocation.objects.update_or_create(
                location_id=structure.get('structure_id'),
                defaults={
                    "location_name": structure.get('name'),
                    "system_id": structure.get('system_id'),
                }
            )
        else:
            try:
                structure_info = fetch_location_name(structure.get(
                    'structure_id'), 'solar_system', token.character_id)

            except Exception:
                structure_info = False
            if structure_info:
                structure_info.save()
                name = structure_info.location_name

        try:
            structure_ob, created = _structures_db_update(
                _corporation,
                structure,
                name
            )

        except MultipleObjectsReturned:
            id_of_first = Structure.objects.filter(
                structure_id=structure.get('structure_id')).order_by("id")[0].id
            Structure.objects.filter(structure_id=structure.get(
                'structure_id')).exclude(id=id_of_first).delete()
            structure_ob, created = _structures_db_update(
                _corporation,
                structure,
                name
            )

        structure_ids.append(structure_ob.structure_id)

    Structure.objects.filter(corporation=_corporation).exclude(
        structure_id__in=structure_ids).delete()  # structures die/leave

    _corporation.last_change_structures = timezone.now()
    _corporation.last_update_structures = timezone.now()
    _corporation.save()

    return f"Updated structures for: {_corporation}"


def update_character_logins_from_corp(corp_id):
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Logins for: {}".format(
        audit_corp.corporation))

    req_scopes = ['esi-corporations.track_members.v1',
                  'esi-characters.read_corporation_roles.v1']
    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"
    try:
        tracking_op = providers.esi.client.Corporation.get_corporations_corporation_id_membertracking(
            corporation_id=corp_id)
        tracking = etag_results(tracking_op, token)

        for c in tracking:
            try:
                ca = CharacterAudit.objects.get(
                    character__character_id=c['character_id'])
                ca.last_known_login = c.get('logon_date', None)
                ca.last_known_logoff = c.get('logoff_date', None)
                ca.save()
            except CharacterAudit.DoesNotExist:
                pass

    except NotModifiedError:
        logger.info("CT: No New Logins for: {}".format(
            audit_corp.corporation))
        pass

    all_chars = CharacterAudit.objects.filter(
        character__corporation_id=corp_id)
    all_chars.update(last_update_login=timezone.now())

    return f"Finished Logins for: {audit_corp.corporation}"


def update_corp_assets(corp_id):
    audit_corp = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Assets for: {}".format(
        audit_corp.corporation))

    req_scopes = ['esi-assets.read_corporation_assets.v1',
                  'esi-characters.read_corporation_roles.v1']
    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        return "No Tokens"
    try:
        failed_locations = []

        assets_op = providers.esi.client.Assets.get_corporations_corporation_id_assets(
            corporation_id=corp_id
        )

        assets = etag_results(
            assets_op, token, disable_verification=CorptoolsConfiguration.skip_verify_assets())

        location_names = list(
            EveLocation.objects.all().values_list('location_id', flat=True)
        )
        _current_type_ids = []
        item_ids = []
        items = []
        for item in assets:
            if item.get('type_id') not in _current_type_ids:
                _current_type_ids.append(item.get('type_id'))
            item_ids.append(item.get('item_id'))
            asset_item = CorpAsset(corporation=audit_corp,
                                   blueprint_copy=item.get(
                                       'is_blueprint_copy'),
                                   singleton=item.get('is_singleton'),
                                   item_id=item.get('item_id'),
                                   location_flag=sanitize_location_flag(
                                       item.get('location_flag')
                                   ),
                                   location_id=item.get('location_id'),
                                   location_type=item.get(
                                       'location_type'),
                                   quantity=item.get('quantity'),
                                   type_id=item.get('type_id'),
                                   type_name_id=item.get('type_id')
                                   )
            if item.get('location_id') not in location_names:
                try:
                    if item.get('location_id') not in failed_locations:
                        new_name = fetch_location_name(
                            item.get('location_id'), item.get('location_flag'), token.character_id)
                        if new_name:
                            new_name.save()
                            location_names.append(item.get('location_id'))
                            asset_item.location_name_id = item.get(
                                'location_id')
                        else:
                            failed_locations.append(item.get('location_id'))
                except Exception:
                    pass  # TODO
            else:
                asset_item.location_name_id = item.get('location_id')

            items.append(asset_item)

        EveItemType.objects.create_bulk_from_esi(_current_type_ids)

        delete_query = CorpAsset.objects.filter(
            corporation=audit_corp)  # Flush Assets
        if delete_query.exists():
            # speed and we are not caring about f-keys or signals on these models
            delete_query._raw_delete(delete_query.db)

        CorpAsset.objects.bulk_create(items)
        que = []
        que.append(run_ozone_levels.si(corp_id))
        que.append(build_managed_asset_locations.si(corp_id))
        chain(que).apply_async(priority=7)
    except NotModifiedError:
        logger.info("CT: No New assets for: {}".format(
            audit_corp.corporation))
        pass

    audit_corp.last_update_assets = timezone.now()
    audit_corp.save()

    return f"Finished assets for: {audit_corp.corporation}"


@shared_task(bind=True, base=QueueOnce)
def run_ozone_levels(self, corp_id):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Ozone for: %s" % (_corporation.corporation))
    _structures = Structure.objects.filter(
        type_id=35841, corporation=_corporation)

    for structure in _structures:
        _quantity = \
            CorpAsset.objects.filter(corporation=_corporation, location_id=structure.structure_id,
                                     type_id=16273).aggregate(ozone=Sum('quantity'))['ozone']
        _used = 0

        try:
            last_ozone = BridgeOzoneLevel.objects.filter(
                station_id=structure.structure_id).order_by('-date')[:1][0].quantity
            delta = last_ozone - _quantity
            _used = (delta if _quantity < last_ozone else 0)
        except Exception:
            pass
        try:
            BridgeOzoneLevel.objects.create(
                station_id=structure.structure_id, quantity=_quantity, used=_used)
        except Exception:
            pass  # dont fail for now
    return f"Finished Ozone for: {_corporation.corporation}"


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


@shared_task(bind=True)
def build_managed_asset_locations(self, corp_id):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("Updating Corporate Hangar Locations for: %s" %
                 (_corporation.corporation))

    req_scopes = ['esi-corporations.read_divisions.v1',
                  'esi-characters.read_corporation_roles.v1']
    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)

    divisions = {}
    if token:
        divs = providers.esi.client.Corporation.get_corporations_corporation_id_divisions(
            corporation_id=_corporation.corporation.corporation_id,
            token=token.valid_access_token()
        ).result()
        for d in divs['hangar']:
            divisions[f"CorpSAG{d['division']}"] = d['name']

    existing_locations = set(list(EveLocation.objects.filter(
        managed=True, managed_corp=_corporation).values_list("location_id", flat=True)))
    current = []
    new = []
    updates = []
    # get office folders and name them
    folders = CorpAsset.objects.filter(corporation=_corporation,
                                       location_flag="OfficeFolder"
                                       ).select_related("location_name")
    names = {}

    grp_ids = [
        12,
        340,
        448,
        649,
    ]

    cat_ids = [
        6
    ]

    corp_hangars = {}
    hangar_ids = []
    for i in folders:
        if i.location_name:
            names[i.item_id] = i.location_name.location_name
        else:
            names[i.item_id] = i.location_id
        corp_hangars[i.item_id] = {}
        for j in range(1, 7):

            # Build the Managed Location for hangars
            id = -int(f"{i.item_id}{j}")
            sag = f"CorpSAG{j}"
            name = f"{names[i.item_id]} > {divisions.get(sag, sag)}"
            hangar_ids.append(id)

            # store for later?
            corp_hangars[i.item_id][sag] = (id, name)

            # update or create the location
            EveLocation.objects.update_or_create(
                location_id=id,
                defaults={
                    "location_name": name,
                    "managed": True,
                    "managed_corp": _corporation
                }
            )

            # update all the assets.
            CorpAsset.objects.filter(
                corporation=_corporation,
                location_id=i.item_id,
                location_flag=sag
            ).update(location_name_id=id)
    # get Containers in office folders and assign the names

    containers = CorpAsset.objects.filter(
        Q(type_name__group_id__in=grp_ids) | Q(
            type_name__group__category_id__in=cat_ids)
    ).filter(
        corporation=_corporation,
        singleton=True,
        location_id__in=folders.values_list(
            "item_id", flat=True),
    ).select_related("location_name", "type_name")

    req_scopes = ['esi-assets.read_corporation_assets.v1',
                  'esi-characters.read_corporation_roles.v1']

    req_roles = ['Director']

    token = get_corp_token(corp_id, req_scopes, req_roles)
    if token:
        ids = list(set(list(containers.values_list("item_id", flat=True))))
        for c in chunks(ids, 250):
            items = providers.esi.client.Assets.post_corporations_corporation_id_assets_names(
                item_ids=c,
                corporation_id=_corporation.corporation.corporation_id,
                token=token.valid_access_token()
            ).result()
            for n in items:
                names[n['item_id']] = n['name']

    # each container gets a location name Station > Hangar > Type
    for c in containers:
        current.append(c.item_id)
        name = names.get(c.item_id, c.type_name.name)
        hangar = divisions.get(c.location_flag, c.location_flag)
        station = names.get(c.location_id, c.location_id)
        item_loc = f"{station} > {hangar} > {name}"
        loc = EveLocation(
            location_id=c.item_id,
            location_name=item_loc,
            managed=True,
            managed_corp=_corporation
        )
        if c.item_id in existing_locations:
            updates.append(loc)
        else:
            new.append(loc)

    if len(new):
        EveLocation.objects.bulk_create(new)

    if len(updates):
        EveLocation.objects.bulk_update(
            updates, ["location_name", "managed", "managed_corp"])

    EveLocation.objects.filter(managed=True, managed_corp=_corporation).exclude(
        location_id__in=current + hangar_ids).delete()

    CorpAsset.objects.filter(location_id__in=current).update(
        location_name_id=F("location_id"))


@shared_task(bind=True, base=QueueOnce)
def build_jb_network(self):
    structures = Structure.objects.filter(type_id=35841).select_related(
        "corporation__corporation", "system_name"
    ).prefetch_related('structureservice_set')

    # second_systems = set()
    output = {}
    regex = r"^(.*) » ([^ - ]*) - (.*)"
    for s in structures:
        matches = re.findall(regex, s.name)
        matches = matches[0]
        output[matches[0]] = {}
        output[matches[0]]["start"] = {"system": s.system_name,
                                       "structure_id": s.structure_id,
                                       "owner": s.corporation.corporation.alliance.alliance_id,
                                       "name": s.name}
        _exit = MapSystem.objects.get(name=matches[1])
        output[matches[0]]["end"] = {"system": _exit,
                                     "name": _exit.name}

    new_jbs = []
    for k, m in output.items():
        new_jbs.append(
            MapJumpBridge(
                structure_id=m["start"]["structure_id"],
                from_solar_system=m["start"]["system"],
                to_solar_system=m["end"]["system"],
                owner_id=m["start"]["owner"]
            )
        )

    MapJumpBridge.objects.filter(manually_input=False).delete()
    MapJumpBridge.objects.bulk_create(new_jbs)

    return f"Created {len(new_jbs)} new JB links"


def update_corporate_contracts(corp_id, force_refresh=False):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)
    logger.debug("updating corporate contracts for: %s" %
                 str(_corporation.corporation.corporation_name))

    req_scopes = ['esi-contracts.read_corporation_contracts.v1',
                  'esi-characters.read_corporation_roles.v1']

    token = get_corp_token(corp_id, req_scopes, False)

    new_contract_ids = []

    if not token:
        return False, []
    try:
        contracts_op = providers.esi.client.Contracts.get_corporations_corporation_id_contracts(
            corporation_id=corp_id)

        contracts = etag_results(
            contracts_op, token, force_refresh=force_refresh)

        contract_models_new = []
        contract_models_old = []
        eve_names = set()
        contract_ids = list(CorporateContract.objects.filter(
            corporation=_corporation).values_list("contract_id", flat=True))
        for c in contracts:  # update labels
            _contract_item = CorporateContract(
                id=CorporateContract.build_pk(
                    _corporation.id, c.get('contract_id')),
                corporation=_corporation,
                assignee_id=c.get('assignee_id'),
                assignee_name_id=c.get('assignee_id'),
                acceptor_id=c.get('acceptor_id'),
                acceptor_name_id=c.get('acceptor_id'),
                contract_id=c.get('contract_id'),
                availability=c.get('availability'),
                buyout=c.get('buyout'),
                collateral=c.get('collateral'),
                date_accepted=c.get('date_accepted'),
                date_completed=c.get('date_completed'),
                date_expired=c.get('date_expired'),
                date_issued=c.get('date_issued'),
                days_to_complete=c.get('days_to_complete'),
                end_location_id=c.get('end_location_id'),
                for_corporation=c.get('for_corporation'),
                issuer_corporation_name_id=c.get('issuer_corporation_id'),
                issuer_name_id=c.get('issuer_id'),
                price=c.get('price'),
                reward=c.get('reward'),
                start_location_id=c.get('start_location_id'),
                status=c.get('status'),
                title=c.get('title'),
                contract_type=c.get('type'),
                volume=c.get('volume')
            )

            eve_names.add(c.get('assignee_id'))
            eve_names.add(c.get('issuer_corporation_id'))
            eve_names.add(c.get('issuer_id'))
            eve_names.add(c.get('acceptor_id'))

            if c.get('contract_id') in contract_ids:
                contract_models_old.append(_contract_item)
            else:
                contract_models_new.append(_contract_item)

        EveName.objects.create_bulk_from_esi(list(eve_names))

        if len(contract_models_new) > 0:
            CorporateContract.objects.bulk_create(
                contract_models_new, batch_size=1000, ignore_conflicts=True)

        if len(contract_models_old) > 0:
            CorporateContract.objects.bulk_update(contract_models_old, batch_size=1000,
                                                  fields=['date_accepted', 'date_completed', 'acceptor_id', 'date_expired', 'status',
                                                          'assignee_name', 'acceptor_name', 'issuer_corporation_name', 'issuer_name'])

        new_contract_ids = [c.contract_id for c in contract_models_new]
        if force_refresh:
            new_contract_ids += [c.contract_id for c in contract_models_old]

    except NotModifiedError:
        logger.info("CT: No New Contracts for: {}".format(
            _corporation.corporation.corporation_name))
        pass

    _corporation.last_update_contracts = timezone.now()
    _corporation.save()

    return "CT: Completed corporate contracts for: %s" % str(_corporation.corporation.corporation_name), new_contract_ids


@shared_task(bind=True, base=QueueOnce)
def update_corporate_contract_items(self, corp_id, contract_id, force_refresh=False):
    _corporation = CorporationAudit.objects.get(
        corporation__corporation_id=corp_id)

    logger.debug("updating contract items for: %s (%s)" %
                 (str(_corporation.corporation.corporation_name), str(contract_id)))

    contract = CorporateContract.objects.get(
        corporation=_corporation,
        contract_id=contract_id)

    if contract.status != "deleted":
        req_scopes = ['esi-contracts.read_corporation_contracts.v1',
                      'esi-characters.read_corporation_roles.v1']

        token = get_corp_token(corp_id, req_scopes, False)

        if not token:
            return False
        try:
            contracts_op = providers.esi.client.Contracts.get_corporations_corporation_id_contracts_contract_id_items(
                corporation_id=corp_id,
                contract_id=contract_id)

            try:
                contracts = etag_results(
                    contracts_op, token, force_refresh=force_refresh)
            except NotModifiedError as e:
                raise e
            except Exception as e:
                if e.status_code == 404:
                    logger.warning("CT: Contract items {} ({}) NOT FOUND ERROR".format(
                        str(_corporation.corporation.corporation_name), str(contract_id)))
                    return
                else:
                    self.retry()

            contract_models_new = []
            _types = set()
            for c in contracts:  # update labels
                _contract_item = CorporateContractItem(
                    contract=contract,
                    is_included=c.get('is_included'),
                    is_singleton=c.get('is_singleton'),
                    quantity=c.get('quantity'),
                    raw_quantity=c.get('raw_quantity'),
                    record_id=c.get('record_id'),
                    type_name_id=c.get('type_id'),
                )
                _types.add(c.get('type_id'))
                contract_models_new.append(_contract_item)
            EveItemType.objects.create_bulk_from_esi(list(_types))
            CorporateContractItem.objects.bulk_create(
                contract_models_new, batch_size=1000, ignore_conflicts=True)
        except NotModifiedError:
            logger.info(
                f"CT: No New Corporate Contract items {str(_corporation.corporation.corporation_name)} ({str(contract_id)})")
            pass

    return f"CT: Completed Corporate Contract items {str(_corporation.corporation.corporation_name)} ({str(contract_id)})"



from ninja import NinjaAPI
from ninja.security import django_auth

from django.conf import settings

from allianceauth.services.hooks import get_extension_logger

from . import character, core, corporation, extras

logger = get_extension_logger(__name__)

api = NinjaAPI(title="CorpTools API", version="0.0.1",
               urls_namespace='corptools:new_api', auth=django_auth, csrf=True,
               openapi_url=settings.DEBUG and "/openapi.json" or "")

# Add the core endpoints
core.setup(api)

# Add the character endpoints
character.setup(api)

# Add the corp endpoints
corporation.setup(api)

# Add the extra endpoints
extras.setup(api)

# Done?



from typing import List

from ninja import NinjaAPI

from django.utils.translation import gettext as _

from corptools import models
from corptools.api import schema
from corptools.api.helpers import get_alts_queryset, get_main_character


class CloneApiEndpoints:

    tags = ["Account"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "account/{character_id}/clones",
            response={200: List[schema.CharacterClones], 403: str},
            tags=self.tags
        )
        def get_character_clones(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            jump_clones = models.JumpClone.objects\
                .filter(character__character__in=characters)\
                .select_related('character__character', 'location_name').prefetch_related('implant_set', 'implant_set__type_name')
            clones = models.Clone.objects\
                .filter(character__character__in=characters)\
                .select_related('character__character', 'location_name')

            table_data = {}
            for char in characters:
                table_data[char.character_name] = {
                    "character": char,
                    "clones": [],
                    "home": None,
                    "last_station_change": None,
                    "last_clone_jump": None
                }

            for j in jump_clones:
                implants = []
                for i in j.implant_set.all():
                    implants.append({
                        "id": i.type_name_id,
                        "name": i.type_name.name
                    })
                loc = None
                if j.location_id:
                    loc = {"id": j.location_id,
                           "name": f"Location ID {j.location_id}"}

                    if j.location_name:
                        loc["name"] = j.location_name.location_name

                table_data[j.character.character.character_name]["clones"].append({
                    "name": j.name,
                    "location": loc,
                    "implants": implants
                }
                )

            for c in clones:
                loc = {"id": c.location_id,
                       "name": f"ID#{c.location_id}"}

                if c.location_name:
                    loc["name"] = c.location_name.location_name

                table_data[c.character.character.character_name]["home"] = loc
                table_data[c.character.character.character_name]["last_station_change"] = c.last_station_change_date
                table_data[c.character.character.character_name]["last_clone_jump"] = c.last_clone_jump_date

            return list(table_data.values())



from datetime import timedelta

from ninja import Field, Schema
from ninja.pagination import LimitOffsetPagination
from ninja.types import DictStrAny

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F, QuerySet, Sum
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

from .. import app_settings, models

logger = get_extension_logger(__name__)


class Paginator(LimitOffsetPagination):
    class Input(Schema):
        limit: int = Field(app_settings.CT_PAGINATION_SIZE, ge=1)
        offset: int = Field(0, ge=0)

    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Input,
        **params: DictStrAny,
    ) -> any:
        offset = pagination.offset
        limit: int = pagination.limit
        return {
            "items": queryset[offset: offset + limit],
            "count": self._items_count(queryset),
        }


def get_main_character(request, character_id):
    perms = True
    main_char = EveCharacter.objects\
        .select_related('character_ownership', 'character_ownership__user__profile', 'character_ownership__user__profile__main_character', )\
        .get(character_id=character_id)
    try:
        main_char = main_char.character_ownership.user.profile.main_character
    except (ObjectDoesNotExist):
        pass

    # check access
    visible = models.CharacterAudit.objects.visible_eve_characters(
        request.user)
    if main_char not in visible:
        account_chars = request.user.profile.main_character.character_ownership.user.character_ownerships.all(
        )
        logger.warning(
            f"{request.user} Can See {list(visible)}, requested {main_char.id}")
        if main_char in account_chars:
            pass
        else:
            perms = False

    if not request.user.has_perm('corptools.view_characteraudit'):
        logger.warning(
            f"{request.user} does not have Perm requested, Requested {main_char.id}")
        perms = False

    return perms, main_char


def get_alts_queryset(main_char):
    try:
        linked_characters = main_char.character_ownership.user.character_ownerships.all(
        ).values_list('character_id', flat=True)

        return EveCharacter.objects.filter(id__in=linked_characters)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def get_corporation_characters(request, corporation_id):
    return EveCharacter.objects.filter(
        character_ownership__user__profile__main_character__corporation_id=corporation_id,
        character_id__in=models.CharacterAudit.objects.visible_eve_characters(
            request.user
        ).values_list("character_id", flat=True)
    )


def round_or_null(value, digits=2):
    if value:
        return round(value, digits)
    else:
        return value


def wallet_check(characters, types, first_parties=None, minimum_amount=None, look_back=30):
    start_date = timezone.now() - timedelta(days=look_back)
    if isinstance(types, str):
        types = [types]

    if isinstance(first_parties, str):
        first_parties = [first_parties]

    qry = models.CharacterWalletJournalEntry.objects.filter(
        character__character__in=characters,
        ref_type__in=types,
        date__gte=start_date)

    if first_parties:
        qry = qry.filter(
            first_party_id__in=first_parties
        )

    if minimum_amount:
        qry = qry.filter(
            amount__gte=minimum_amount
        )

    return qry


def mining_check(characters, groups, look_back=30):
    start_date = timezone.now() - timedelta(days=look_back)
    return models.CharacterMiningLedger.objects.filter(
        character__character__in=characters,
        type_name__group_id__in=groups,
        date__gte=start_date
    )


def glance_incursion_check(characters):
    from_ids = [1000125]
    types = ["corporate_reward_payout"]
    return wallet_check(characters, types, first_parties=from_ids).aggregate(total=Sum("amount"))["total"]


def glances_missions_check(characters):
    types = ["agent_mission_reward", "agent_mission_time_bonus_reward"]
    return wallet_check(characters, types).aggregate(total=Sum("amount"))["total"]


def glances_ratting_check(characters):
    types = ["bounty_prizes"]
    min_amount = 5000000
    # 5 mill ticks should cover gate rats etc. but still show passive ratting
    return wallet_check(characters, types, minimum_amount=min_amount).aggregate(total=Sum("amount"))["total"]


def glances_pochven_check(characters):
    from_ids = [1000298]
    types = ["corporate_reward_payout"]
    return wallet_check(characters, types, first_parties=from_ids).aggregate(total=Sum("amount"))["total"]


def glances_market_check(characters):
    types = ["brokers_fee", "market_provider_tax"]
    return wallet_check(characters, types).exists()


def glances_industry_check(characters):
    types = ["industry_job_tax"]
    return wallet_check(characters, types).exists()


def glances_pi_check(characters):
    types = ["planetary_import_tax",
             "planetary_export_tax", "planetary_construction"]
    return wallet_check(characters, types).exists()


def glances_ore_check(characters):
    group_ids = [
        450,
        451,
        452,
        453,
        454,
        455,
        456,
        457,
        458,
        459,
        460,
        461,
        462,
        467,
        468,
        469,
        2024,
        4029,
        4030,
        4031,
        4513,
        4514,
        4515,
        4516,
        4568
    ]
    return mining_check(
        characters,
        group_ids
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def glances_moon_check(characters):
    group_ids = [
        1884,
        1920,
        1921,
        1922,
        1920
    ]
    return mining_check(
        characters,
        group_ids
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def glances_ice_check(characters):
    group_ids = [
        465,
        903
    ]
    return mining_check(
        characters,
        group_ids
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def glances_gas_check(characters):
    group_ids = [711]
    return mining_check(
        characters,
        group_ids
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def roundFloat(input):
    if input:
        return int(input)
    else:
        return input



# Generated by Django 4.0.8 on 2023-01-12 09:04

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0081_characterlocation_current_ship_name_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='characteraudit',
            name='last_update_contracts',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.CreateModel(
            name='Contract',
            fields=[
                ('id', models.CharField(max_length=50,
                 primary_key=True, serialize=False)),
                ('contract_id', models.IntegerField()),
                ('acceptor_id', models.IntegerField()),
                ('assignee_id', models.IntegerField()),
                ('issuer_id', models.IntegerField()),
                ('issuer_corporation_id', models.IntegerField()),
                ('collateral', models.DecimalField(blank=True,
                 decimal_places=2, max_digits=20, null=True)),
                ('buyout', models.DecimalField(blank=True,
                 decimal_places=2, max_digits=20, null=True)),
                ('price', models.DecimalField(blank=True,
                 decimal_places=2, max_digits=20, null=True)),
                ('reward', models.DecimalField(blank=True,
                 decimal_places=2, max_digits=20, null=True)),
                ('volume', models.DecimalField(blank=True,
                 decimal_places=2, max_digits=20, null=True)),
                ('days_to_complete', models.IntegerField(blank=True, null=True)),
                ('start_location_id', models.BigIntegerField(blank=True, null=True)),
                ('end_location_id', models.BigIntegerField(blank=True, null=True)),
                ('for_corporation', models.BooleanField()),
                ('date_accepted', models.DateTimeField(blank=True, null=True)),
                ('date_completed', models.DateTimeField(blank=True, null=True)),
                ('date_expired', models.DateTimeField()),
                ('date_issued', models.DateTimeField()),
                ('status', models.CharField(max_length=25)),
                ('contract_type', models.CharField(max_length=20)),
                ('availability', models.CharField(max_length=15)),
                ('title', models.CharField(max_length=256)),
                ('acceptor_name', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                 related_name='contractacceptor', to='corptools.evename')),
                ('assignee_name', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                 related_name='contractassignee', to='corptools.evename')),
                ('character', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.characteraudit')),
                ('end_location_name', models.ForeignKey(default=None, null=True,
                 on_delete=django.db.models.deletion.SET_NULL, related_name='contractto', to='corptools.evelocation')),
                ('issuer_corporation_name', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                 related_name='contractissuercorporation', to='corptools.evename')),
                ('issuer_name', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                 related_name='contractissuer', to='corptools.evename')),
                ('start_location_name', models.ForeignKey(default=None, null=True,
                 on_delete=django.db.models.deletion.SET_NULL, related_name='contractfrom', to='corptools.evelocation')),
            ],
            options={
                'unique_together': {('character', 'contract_id')},
            },
        ),
        migrations.CreateModel(
            name='ContractItem',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('is_included', models.BooleanField()),
                ('is_singleton', models.BooleanField()),
                ('quantity', models.IntegerField()),
                ('raw_quantity', models.IntegerField(blank=True, null=True)),
                ('record_id', models.BigIntegerField()),
                ('contract', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.contract')),
                ('type_name', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.eveitemtype')),
            ],
            options={
                'unique_together': {('record_id', 'contract')},
            },
        ),
    ]



# Generated by Django 4.2.13 on 2024-05-27 10:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0107_rename_structures_homestationfilter_evelocation'),
    ]

    operations = [
        migrations.CreateModel(
            name='JumpCloneFilter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('description', models.CharField(max_length=500)),
                ('evelocation', models.ManyToManyField(blank=True, help_text='Limit filter to specific Structures', to='corptools.evelocation')),
            ],
            options={
                'verbose_name': 'Smart Filter: Jump Clone',
                'verbose_name_plural': 'Smart Filter: Jump Clone',
            },
        ),
    ]
